
# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict

from copy import copy
import functools

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor
from leaf_common.asyncio.asyncio_executor_pool import AsyncioExecutorPool
from leaf_server_common.logging.logging_setup import setup_extra_logging_fields

from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.message_journal import MessageJournal
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.origination import Origination


# pylint: disable=too-many-instance-attributes
class SessionInvocationContext(InvocationContext):
    """
    Implementation of InvocationContext which encapsulates specific policy classes that pertain to
    a single invocation of an AgentSession, whether by way of a
    service call or library call.
    """

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def __init__(self, agent_name: str,
                 async_session_factory: AsyncAgentSessionFactory,
                 async_executors_pool: AsyncioExecutorPool,
                 llm_factory: ContextTypeLlmFactory,
                 toolbox_factory: ContextTypeToolboxFactory = None,
                 metadata: Dict[str, str] = None,
                 reservationist: Reservationist = None,
                 port: int = None,
                 effective_invocation: str = "chatbot"):
        """
        Constructor

        :param agent_name: The name of the agent
        :param async_session_factory: The AsyncAgentSessionFactory to use
                        when connecting with external agents.
        :param async_executors_pool: pool of AsyncioExecutors to use for obtaining
                         an executor instance to use for this context;
        :param llm_factory: The ContextTypeLlmFactory instance
        :param toolbox_factory: The ContextTypeToolboxFactory instance
        :param metadata: A request metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        :param reservationist: The Reservationist instance to use.
        :param port: The port on which the server was started
        :param effective_invocation: A string representing the effective invocation of the session.
        """

        # From args
        self.agent_name: str = agent_name
        self.async_session_factory: AsyncAgentSessionFactory = async_session_factory
        self.async_executors_pool: AsyncioExecutorPool = async_executors_pool
        self.llm_factory: ContextTypeLlmFactory = llm_factory
        self.toolbox_factory: ContextTypeToolboxFactory = toolbox_factory
        self.metadata: Dict[str, str] = metadata
        self.reservationist: Reservationist = reservationist
        self.port: int = port
        self.effective_invocation: str = effective_invocation

        # Internal
        # Get an async executor to run all tasks for this session instance:
        self.asyncio_executor: AsyncioExecutor = self.async_executors_pool.get_executor()
        self.request_reporting: Dict[str, Any] = {}
        self.origination: Origination = Origination()

        # Anything that has to do with the queue will need a new instance in
        # safe_shallow_copy() below to keep AsyncDirectAgentSessions happy.
        self.queue: AsyncCollatingQueue = AsyncCollatingQueue()
        self.journal: Journal = MessageJournal(self.queue)

    def start(self):
        """
        Starts the active components of this invocation context.
        Do this separately from constructor for more control.
        Currently, we only start internal AsyncioExecutor.
        It could be already running, but starting it twice is allowed.
        """
        # Wrap it up into a single function with no parameters
        # for easier handling downstream.
        logging_setup: Callable = functools.partial(setup_extra_logging_fields, metadata_dict=self.metadata)
        self.asyncio_executor.start()
        # Run logging setup as event-loop initialization step -
        # make sure it is finished before we start to use this AsyncioExecutor instance.
        self.asyncio_executor.initialize(logging_setup)

    def get_effective_invocation(self) -> str:
        """
        :return: The effective invocation of the session
        """
        return self.effective_invocation

    def get_async_session_factory(self) -> AsyncAgentSessionFactory:
        """
        :return: The AsyncAgentSessionFactory associated with the invocation
        """
        return self.async_session_factory

    def get_asyncio_executor(self) -> AsyncioExecutor:
        """
        :return: The AsyncioExecutor associated with the invocation
        """
        return self.asyncio_executor

    def get_origination(self) -> Origination:
        """
        :return: The Origination instance carrying state about tool instantation
                during the course of the AgentSession.
        """
        return self.origination

    def get_journal(self) -> Journal:
        """
        :return: The Journal instance that allows message reporting
                during the course of the AgentSession.
        """
        return self.journal

    def get_queue(self) -> AsyncCollatingQueue:
        """
        :return: The AsyncCollatingQueue instance via which messages are streamed to the
                AgentSession mechanics
        """
        return self.queue

    def get_metadata(self) -> Dict[str, str]:
        """
        :return: The metadata to pass along with any request
        """
        return self.metadata

    def close(self):
        """
        Release resources owned by this context
        """
        if self.asyncio_executor is not None:
            self.async_executors_pool.return_executor(self.asyncio_executor)
            self.asyncio_executor = None
        if self.queue is not None:
            self.queue.close()

    def get_request_reporting(self) -> Dict[str, Any]:
        """
        :return: The request reporting dictionary
        """
        return self.request_reporting

    def get_llm_factory(self) -> ContextTypeLlmFactory:
        """
        :return: The ContextTypeLlmFactory instance for the session
        """
        return self.llm_factory

    def get_toolbox_factory(self) -> ContextTypeToolboxFactory:
        """
        :return: The ContextTypeToolboxFactory instance for the session
        """
        return self.toolbox_factory

    def get_reservationist(self) -> Reservationist:
        """
        :return: The Reservationist instance for the session
        """
        return self.reservationist

    def get_agent_name(self) -> str:
        """
        :return: The agent name for the session
        """
        return self.agent_name

    def get_port(self) -> int:
        """
        :return: The port on which the server was started
        """
        return self.port

    def reset(self):
        """
        Resets the instance for a subsequent use for another exchange with the agent network.
        """
        # Origination needs to be reset so that origin information can match up
        # with what is in the chat_context. If we do not reset this, then library calls
        # to DirectAgentSession do not properly carry forward any memory of the conversation
        # in subsequent interactions with the same network.
        self.origination.reset()
        self.queue.reset()

    def safe_shallow_copy(self, invocation: str = None) -> SessionInvocationContext:
        """
        Makes a safe shallow copy of the invocation context.
        Generally used with direct sessions.

        :param invocation: String describing how the agent wants to be invoked.
                            Can be: "chatbot" - implies waiting for an answer.
                                    "event" - implies no answer needed
                                    None - implies chatbot
        :return: A copy of the invocation context
        """

        invocation_context: SessionInvocationContext = copy(self)

        # We need a different queue in order to call external agents with direct sessions.
        invocation_context.queue: AsyncCollatingQueue = AsyncCollatingQueue()

        # Now that the queue has changed, we need a new Journal as well
        # to be sure that the messages are sent to the correct queue.
        invocation_context.journal: Journal = MessageJournal(invocation_context.queue)

        # If an invocation was provided, use it
        if invocation is not None:
            invocation_context.effective_invocation = invocation

        return invocation_context




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Tuple

from neuro_san.internals.interfaces.startable import Startable


class RegistryObserver(Startable):
    """
    Interface for specific kinds of filesystem observing
    """

    def start(self):
        """
        Start running observer
        """
        raise NotImplementedError

    def reset_event_counters(self) -> Tuple[int, int, int]:
        """
        Reset event counters and return current counters.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Set

from logging import getLogger
from logging import Logger

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.internals.validation.network.abstract_network_validator import AbstractNetworkValidator


class UnreachableNodesNetworkValidator(AbstractNetworkValidator):
    """
    AbstractNetworkValidator that looks for topological issues in an agent network.
    Specifically, unreachable nodes or issues with number of front men.
    """

    def __init__(self):
        """
        Constructor
        """
        self.logger: Logger = getLogger(self.__class__.__name__)

    def validate_name_to_spec_dict(self, name_to_spec: Dict[str, Any]) -> List[str]:
        """
        Validate the agent network, specifically in the form of a name -> agent spec dictionary.

        :param name_to_spec: The name -> agent spec dictionary to validate
        :return: A list of error messages
        """
        errors: List[str] = []

        self.logger.info("Validating agent network structure...")

        # Find top agents
        top_agents: Set[str] = self.find_all_top_agents(name_to_spec)

        if len(top_agents) == 0:
            errors.append("No top agent found in network")
        elif len(top_agents) > 1:
            errors.append(f"Multiple top agents found: {sorted(top_agents)}. Expected exactly one.")

        # Find unreachable agents (only meaningful if we have exactly one top agent)
        unreachable_agents: Set[str] = set()
        if len(top_agents) == 1:
            top_agent: str = next(iter(top_agents))
            unreachable_agents = self.find_unreachable_agents(name_to_spec, top_agent)
            if unreachable_agents:
                errors.append(f"Unreachable agents found: {sorted(unreachable_agents)}")

        if len(errors) > 0:
            # Only warn if there is a problem
            self.logger.warning(str(errors))

        return errors

    def find_all_top_agents(self, name_to_spec: Dict[str, Any]) -> Set[str]:
        """
        Find all top agents - agents that have down-chains but are not down-chains of others.

        :param name_to_spec: The agent network to validate
        :return: Set of top agent names
        """
        all_down_chains: Set[str] = set()
        has_down_chains: Set[str] = set()

        for agent_name, agent_config in name_to_spec.items():
            down_chains: List[str] = agent_config.get("tools", [])
            if down_chains:

                has_down_chains.add(agent_name)

                safe_down_chains: List[str] = self.remove_dictionary_tools(down_chains)
                all_down_chains.update(safe_down_chains)

        # Potential top agents are agents that have down-chains but are not down-chains of others
        top_agents: Set[str] = has_down_chains - all_down_chains

        # Special case: If there's only one agent in the network, it's always a top agent
        if len(top_agents) == 0 and len(name_to_spec) == 1:
            # It's OK to have a single top agent with no down-chains
            one_top: str = list(name_to_spec.keys())[0]
            top_agents.add(one_top)

        return top_agents

    def find_unreachable_agents(self, name_to_spec: Dict[str, Any], top_agent: str) -> Set[str]:
        """
        Find agents that are unreachable from the top agent using Depth-First Search (DFS) traversal.

        :param name_to_spec: The agent network to validate
        :param top_agent: The single top agent to start from
        :return: Set of unreachable agent names
        """
        # Step 1: Initialize set to track all agents we can reach from top agent
        reachable_agents: Set[str] = set()

        # Step 2: Initialize visited set to track DFS traversal (prevents infinite loops in cycles)
        visited: Set[str] = set()

        # Step 3: Start DFS traversal from the top agent to find all reachable agents
        self.dfs_reachability_traversal(name_to_spec, top_agent, visited, reachable_agents)

        # Step 4: Get complete set of all agents in the network
        all_agents: Set[str] = set(name_to_spec.keys())

        # Step 5: Calculate unreachable agents by subtracting reachable from all agents
        unreachable_agents: Set[str] = all_agents - reachable_agents

        # Step 6: Return the set of agents that cannot be reached from top agent
        return unreachable_agents

    def dfs_reachability_traversal(self, name_to_spec: Dict[str, Any], agent: str,
                                   visited: Set[str], reachable_agents: Set[str]):
        """
        Perform DFS traversal to find all agents reachable from a specific starting agent.

        :param name_to_spec: The agent network to validate
        :param agent: Current agent being visited
        :param visited: Set of agents already visited in this traversal (prevents infinite loops)
        :param reachable_agents: Set to collect all agents that can be reached
        """
        # Step 1: Check if we've already visited this agent or if it doesn't exist in network
        if agent in visited or agent not in name_to_spec:
            return  # Skip already visited agents or non-existent agents

        # Step 2: Mark current agent as visited to prevent revisiting
        visited.add(agent)

        # Step 3: Add current agent to our reachable set
        reachable_agents.add(agent)

        # Step 4: Get all child agents (down_chains) of current agent
        empty: Dict[str, Any] = {}
        no_tools: List[str] = []

        agent_spec: Dict[str, Any] = name_to_spec.get(agent, empty)
        extractor = DictionaryExtractor(agent_spec)

        traditional_down_chains: List[str] = extractor.get("tools", no_tools)
        coded_tool_down_chains: List[str] = list(extractor.get("args.tools", empty).values())
        down_chains: List[str] = traditional_down_chains + coded_tool_down_chains
        safe_down_chains: List[str] = self.remove_dictionary_tools(down_chains)

        # Step 5: Recursively visit each child agent to continue the traversal
        for child_agent in safe_down_chains:
            # Skip URL/path tools - they're not agents
            if not self.is_url_or_path(child_agent):
                # Visit each child - the recursion will handle visited check and network existence
                self.dfs_reachability_traversal(name_to_spec, child_agent, visited, reachable_agents)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""

from neuro_san.service.mcp.interfaces.client_session_policy import ClientSessionPolicy
from neuro_san.service.mcp.interfaces.client_session import ClientSession


class McpNoSessionsPolicy(ClientSessionPolicy):
    """
    Policy class for scenario when client sessions are not supported by the MCP service.
    """

    def create_session(self) -> ClientSession:
        """
        Create a None client session if client sessions are not supported.
        :return: None
        """
        return None

    def activate_session(self, session_id: str) -> bool:
        """
        For "no sessions" policy, always return True.
        """
        return True

    def delete_session(self, session_id: str) -> bool:
        """
        For "no sessions" policy, always return True.
        """
        return True

    def is_session_active(self, session_id: str) -> bool:
        """
        For "no sessions" policy, always return True.
        """
        return True




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""

from typing import Any
from typing import Dict
from typing import List

import json
import random
import threading

import tornado

from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.interfaces.agent_state_listener import AgentStateListener
from neuro_san.internals.interfaces.agent_storage_source import AgentStorageSource
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.service.generic.agent_server_logging import AgentServerLogging
from neuro_san.service.generic.async_agent_service_provider import AsyncAgentServiceProvider
from neuro_san.service.http.config.http_server_config import HttpServerConfig
from neuro_san.service.http.handlers.concierge_handler import ConciergeHandler
from neuro_san.service.http.handlers.connectivity_handler import ConnectivityHandler
from neuro_san.service.http.handlers.function_handler import FunctionHandler
from neuro_san.service.http.handlers.health_check_handler import HealthCheckHandler
from neuro_san.service.http.handlers.openapi_publish_handler import OpenApiPublishHandler
from neuro_san.service.http.handlers.streaming_chat_handler import StreamingChatHandler
from neuro_san.service.http.logging.http_logger import HttpLogger
from neuro_san.service.http.server.agent_authorization_policy import AgentAuthorizationPolicy
from neuro_san.service.http.server.http_server_app import HttpServerApp
from neuro_san.service.http.server.resources_usage_logger import ResourcesUsageLogger
from neuro_san.service.interfaces.agent_authorizer import AgentAuthorizer
from neuro_san.service.interfaces.agent_server import AgentServer
from neuro_san.service.interfaces.event_loop_logger import EventLoopLogger
from neuro_san.internals.interfaces.startable import Startable
from neuro_san.service.mcp.handlers.mcp_root_handler import McpRootHandler
from neuro_san.service.utils.server_context import ServerContext
from neuro_san.service.utils.server_status import ServerStatus


DEFAULT_SERVER_NAME: str = 'neuro-san.Agent'
DEFAULT_SERVER_NAME_FOR_LOGS: str = 'Agent Server'
DEFAULT_MAX_CONCURRENT_REQUESTS: int = 10

# Better that we kill ourselves than kubernetes doing it for us
# in the middle of a request if there are resource leaks.
# This is per the lifetime of the server (before it kills itself).
DEFAULT_REQUEST_LIMIT: int = 1000 * 1000


class HttpServer(AgentStateListener):
    """
    Class provides simple http endpoint for neuro-san API.
    """
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments, too-many-positional-arguments

    TIMEOUT_TO_START_SECONDS: int = 10

    def __init__(self,
                 server_context: ServerContext,
                 server_config: HttpServerConfig,
                 openapi_service_spec_path: str,
                 requests_limit: int,
                 logging_config: Dict[str, Any],
                 forwarded_request_metadata: str = AgentServer.DEFAULT_FORWARDED_REQUEST_METADATA):
        """
        Constructor:
        :param server_context: ServerContext with global-ish state
        :param server_config: http server run-time configuration
        :param openapi_service_spec_path: path to a file with OpenAPI service specification;
        :param requests_limit: The number of requests to service before shutting down.
                        This is useful to be sure production environments can handle
                        a service occasionally going down.
        :param logging_config: logging configuration
        :param forwarded_request_metadata: A space-delimited list of http metadata request keys
               to forward to logs/other requests
        """
        self.server_name_for_logs: str = "Http Server"
        self.server_config = server_config
        self.http_port = self.server_config.http_port
        self.server_context: ServerContext = server_context
        self.logging_config: Dict[str, Any] = logging_config

        # Randomize requests limit for this server instance.
        # Lower and upper bounds for number of requests before shutting down
        if requests_limit == -1:
            # Unlimited requests
            self.requests_limit = -1
        else:
            request_limit_lower = round(requests_limit * 0.90)
            request_limit_upper = round(requests_limit * 1.10)
            self.requests_limit = random.randint(request_limit_lower, request_limit_upper)

        self.openapi_service_spec_path: str = openapi_service_spec_path
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata.split(" ")
        self.logger = HttpLogger(self.forwarded_request_metadata, self.logging_config)
        self.allowed_agents: Dict[str, AsyncAgentServiceProvider] = {}
        self.authorization_policy: AgentAuthorizer = AgentAuthorizationPolicy(self.allowed_agents)
        self.lock = threading.Lock()

        # Add listener to handle adding per-agent http service
        # (services map is defined by self.allowed_agents dictionary)
        network_storage_dict: Dict[str, AgentNetworkStorage] = self.server_context.get_network_storage_dict()
        for network_storage in network_storage_dict.values():
            network_storage.add_listener(self)

    def start(self, startables: List[Startable]):
        """
        Method to be called by a thread running tornado HTTP server
        to actually start serving requests.
        :param startables: List of Startable instances to start once server
            has forked its multiple running instances.
        """
        app = self.make_app(self.requests_limit, self.logger)

        self.logger.debug({}, "Serving agents: %s", repr(self.allowed_agents.keys()))

        # Create an HTTP server with run-time parameters
        server = tornado.httpserver.HTTPServer(
            app,
            idle_connection_timeout=self.server_config.http_idle_connection_timeout_seconds
        )

        if self.server_config.http_server_monitor_interval_seconds > 0:
            # Add resources usage logger to list of things we need to start
            # after http server is spun up:
            resources_logger: Startable =\
                ResourcesUsageLogger(
                    self.server_config.http_server_monitor_interval_seconds, self.http_port, self.logger)
            startables.append(resources_logger)

        # Bind the socket with a custom backlog
        server.bind(self.http_port, backlog=self.server_config.http_connections_backlog)

        # Start N child processes (0 = one per CPU core)
        server.start(self.server_config.http_server_instances)

        server_status: ServerStatus = self.server_context.get_server_status()
        server_status.http_service.set_status(True)
        self.logger.info({}, "HTTP server is running %d instances on port %d with backlog %d",
                         self.server_config.http_server_instances,
                         self.http_port,
                         self.server_config.http_connections_backlog)
        self.logger.info({}, "HTTP server idle connections timeout: %d seconds",
                         self.server_config.http_idle_connection_timeout_seconds)
        self.logger.info({}, "HTTP server is shutting down after %d requests", self.requests_limit)

        # If HTTP server is ready, our MCP server is also ready, if requested to run.
        if server_status.mcp_service.is_requested():
            server_status.mcp_service.set_status(True)
            mcp_version: str = self.server_context.get_mcp_server_context().get_protocol_version()
            self.logger.info({}, f"MCP server is running protocol {mcp_version}")

        if startables:
            for startable in startables:
                if isinstance(startable, Startable):
                    try:
                        startable.start()
                    except Exception as exception:  # pylint: disable=broad-exception-caught
                        self.logger.error(
                            {}, "Failed to start %s: %s",
                            startable.__class__.__name__, str(exception))

        tornado.ioloop.IOLoop.current().start()
        self.logger.info({}, "Http server stopped.")

    def make_app(self, requests_limit: int, logger: EventLoopLogger):
        """
        Construct tornado HTTP "application" to run.
        """
        # Do we need to enable HTTP request handlers?
        enable_http_handlers: bool = self.server_context.get_server_status().http_service.is_requested()

        request_initialize_data: Dict[str, Any] = self.build_request_data()
        live_request_initialize_data: Dict[str, Any] = {
            "forwarded_request_metadata": self.forwarded_request_metadata,
            "server_status": self.server_context.get_server_status(),
            "op": "live",
            "logging_config": self.logging_config
        }
        ready_request_initialize_data: Dict[str, Any] = {
            "forwarded_request_metadata": self.forwarded_request_metadata,
            "server_status": self.server_context.get_server_status(),
            "op": "ready",
            "logging_config": self.logging_config
        }
        handlers = []
        # Health check handlers are enabled always
        handlers.append(("/", HealthCheckHandler, ready_request_initialize_data))
        handlers.append(("/healthz", HealthCheckHandler, ready_request_initialize_data))
        handlers.append(("/readyz", HealthCheckHandler, ready_request_initialize_data))
        handlers.append(("/livez", HealthCheckHandler, live_request_initialize_data))

        if enable_http_handlers:
            handlers.append(("/api/v1/list", ConciergeHandler, request_initialize_data))
            handlers.append(("/api/v1/docs", OpenApiPublishHandler, request_initialize_data))

            # Register templated request paths for agent API methods:
            # regexp format used here is that of Python Re standard library.
            handlers.append((r"/api/v1/(.+)/function", FunctionHandler, request_initialize_data))
            handlers.append((r"/api/v1/(.+)/connectivity", ConnectivityHandler, request_initialize_data))
            handlers.append((r"/api/v1/(.+)/streaming_chat", StreamingChatHandler, request_initialize_data))

        # Register MCP "root" handler for all MCP requests
        # if MCP server is enabled:
        if self.server_context.get_mcp_server_context().is_enabled():
            handlers.append((r"/mcp", McpRootHandler, request_initialize_data))

        return HttpServerApp(handlers, requests_limit, logger, self.forwarded_request_metadata)

    def agent_added(self, agent_name: str, source: AgentStorageSource):
        """
        Add agent to the map of known agents
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        agent_network_provider: AgentNetworkProvider = source.get_agent_network_provider(agent_name)
        # There are some timing scenarios where at this point agent network provider might not be available,
        # even though we got a notification that agent is added.
        # One example: temporary agent network expiration.
        # In this case, log the message and finish processing.
        if agent_network_provider is None:
            self.logger.info({}, "Agent %s has already expired - adding to service cancelled.",
                             agent_name)
            return

        # Convert back to a single string as required by constructor
        request_metadata_str: str = " ".join(self.forwarded_request_metadata)
        agent_server_logging: AgentServerLogging = \
            AgentServerLogging(self.server_name_for_logs, request_metadata_str, self.logging_config)
        agent_service_provider: AsyncAgentServiceProvider = \
            AsyncAgentServiceProvider(
                self.logger,
                None,
                agent_name,
                agent_network_provider,
                agent_server_logging,
                self.server_context)
        self.allowed_agents[agent_name] = agent_service_provider
        self.logger.info({}, "Added agent %s to allowed http service list", agent_name)

    def agent_removed(self, agent_name: str, source: AgentStorageSource):
        """
        Remove agent from the map of known agents
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        self.allowed_agents.pop(agent_name, None)
        self.logger.info({}, "Removed agent %s from allowed http service list", agent_name)

    def agent_modified(self, agent_name: str, source: AgentStorageSource):
        """
        Agent is being modified in the service scope.
        :param agent_name: name of an agent
        :param source: The AgentStorageSource source of the message
        """
        agent_service_provider: AsyncAgentServiceProvider = self.allowed_agents.get(agent_name, None)
        if agent_service_provider is not None:
            agent_service_provider.reset_service()
            self.logger.info({}, "Reset service for modified agent %s", agent_name)

    def build_request_data(self) -> Dict[str, Any]:
        """
        Build request data for Http handlers.
        :return: a dictionary with request data to be passed to an http handler.
        """
        open_api_dict: Dict[str, Any] = None
        try:
            with open(self.openapi_service_spec_path, "r", encoding='utf-8') as f_out:
                open_api_dict = json.load(f_out)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(f"Failed to load '{self.openapi_service_spec_path}'") from exc

        return {
            "agent_policy": self.authorization_policy,
            "forwarded_request_metadata": self.forwarded_request_metadata,
            "openapi_service_spec": open_api_dict,
            "server_context": self.server_context,
            "logging_config": self.logging_config
        }




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings


GRPC_GENERATED_VERSION = '1.78.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in neuro_san/api/grpc/chat_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: neuro_san/api/grpc/mime_data.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'neuro_san/api/grpc/mime_data.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"neuro_san/api/grpc/mime_data.proto\x12-dev.cognizant_ai.neuro_san.api.grpc.mime_data\"H\n\x08MimeData\x12\x1c\n\tmime_type\x18\x01 \x01(\tR\tmime_type\x12\x1e\n\nmime_bytes\x18\x02 \x01(\x0cR\nmime_bytesBoZmgithub.com/cognizant-ai-lab/neuro_san/internal/gen/dev.cognizant_ai/neuro_san/api/grpc/mime_data/v1;mime_datab\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'neuro_san.api.grpc.mime_data_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Zmgithub.com/cognizant-ai-lab/neuro_san/internal/gen/dev.cognizant_ai/neuro_san/api/grpc/mime_data/v1;mime_data'
  _globals['_MIMEDATA']._serialized_start=85
  _globals['_MIMEDATA']._serialized_end=157
# @@protoc_insertion_point(module_scope)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""

import contextvars
import logging


class LogContextFilter(logging.Filter):
    """
    Custom logging filter for Http server.
    """

    def filter(self, record):
        """
        Logging filter: add key-value pairs from log_context
        to logging record to be used.
        """
        ctx = LogContextFilter.log_context.get()
        for key, value in ctx.items():
            setattr(record, key, value)
        return True

    @classmethod
    def set_log_context(cls):
        """
        Create log context class instance.
        """
        cls.log_context = contextvars.ContextVar("http_server_context", default={})




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.graph.activations.calling_activation import CallingActivation
from neuro_san.internals.interfaces.front_man import FrontMan
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.run_context.interfaces.run import Run


class FrontManActivation(CallingActivation, FrontMan):
    """
    A CallingActivation implementation which is the root of the call graph.
    """

    async def create_any_resources(self):
        """
        Creates resources that will be used throughout the lifetime of the component.
        """
        await self.create_resources()

    async def submit_message(self, user_input: str) -> List[BaseMessage]:
        """
        Entry-point method for callers of the root of the Activation tree.

        :param user_input: An input string from the user.
        :return: A list of response messages for the run
        """
        # Initialize our return value
        messages: List[BaseMessage] = []

        current_run: Run = await self.run_context.submit_message(user_input)

        terminate = False
        while not terminate:
            if self.run_context is None:
                # Breaking from inside a container during cleanup can yield a None
                # run_context
                break

            current_run = await self.run_context.wait_on_run(current_run, self.journal)

            if current_run.requires_action():
                current_run = await self.make_tool_function_calls(current_run)
            else:
                # Needs to get more information from the user on the basic task
                # of collecting information from the user about the current run.
                if self.run_context is None:
                    # Breaking from inside a container during cleanup can yield a None
                    # run_context
                    break
                messages = await self.run_context.get_response()
                terminate = True

        return messages

    def update_invocation_context(self, invocation_context: InvocationContext):
        """
        Update internal state based on the InvocationContext instance passed in.
        :param invocation_context: The context policy container that pertains to the invocation
        """
        self.journal = invocation_context.get_journal()
        if self.run_context is not None:
            self.run_context.update_invocation_context(invocation_context)

    async def build(self) -> BaseMessage:
        """
        Main entry point to the class.

        :return: A BaseMessage produced during this process.
        """
        # This is never called for a FrontMan, but is needed to satisfy the
        # class heirarchy stemming from CallableActivation.
        # A FrontMan is not Callable.
        raise NotImplementedError

    async def delete_any_resources(self):
        """
        Cleans up after any allocated resources
        """
        await self.delete_resources(None)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Set
from typing import Union

from asyncio import Event
from time import time

from neuro_san.interfaces.reservation import Reservation
from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.interfaces.reservations_storage import ReservationsStorage
from neuro_san.internals.reservations.agent_reservation import AgentReservation


class AbstractAgentReservationist(Reservationist):
    """
    Reservationist implementation that allows procurement of agent ids
    for a specific amount of time for direct sessions only.
    """

    def __init__(self, reservations_storage: Set[ReservationsStorage],
                 max_lifetime_in_seconds: float = Reservationist.DEFAULT_LIFETIME):
        """
        Constructor

        :param reservations_storage: A Set of ReservationsStorage instances that manage
                            any temporary networks
        :param max_lifetime_in_seconds: The maximum lifetime allowed for any Reservation.
        """
        self.reservations_storage: Set[ReservationsStorage] = reservations_storage
        self.max_lifetime_in_seconds: float = max_lifetime_in_seconds

    async def reserve(self, lifetime_in_seconds: float = Reservationist.DEFAULT_LIFETIME,
                      prefix: str = "") -> Reservation:
        """
        Creates a reservation for a specific time period.

        :param lifetime_in_seconds: The lifetime of the reservation
        :param prefix: A string prefix to prepend to the id so as to provide external context.
        :return: A new Reservation object corresponding to the request.
        """
        actual_lifetime: float = min(lifetime_in_seconds, self.max_lifetime_in_seconds)
        reservation = AgentReservation(actual_lifetime, prefix=prefix)
        return reservation

    async def deploy(self,
                     deployment_dict: Dict[Reservation,
                                           Dict[Union[str, Event],
                                                Dict[AgentReservation,
                                                     Dict[str, Any]]]],
                     confirmation: bool = False) -> Event:
        """
        Deploy multiple Reservations.

        :param deployment_dict: A dictionary whose keys are Reservation objects
                    previously obtained with reserve() and whose values are each an
                    object to deploy using the key as Reservation.
                    In this case, the object value to deploy is another dictionary where:
                        keys can be either a string or an asyncio.Event
                        values are a dictionary of AgentReservations -> agent_spec dictionaries
                            That is, these values are deployment_dicts from an AccumulatingAgentReservationist.
        :param confirmation: When True, indicates a confirmation Event is desired.
                            When False (the default), no such confirmation is needed.
                            This is ignored on this instance
        :return: Either a confiramtion Event object whose wait() method will confirm
                that the deployment is complete, or None, depending on the value passed
                for confirmation.
                This instance always returns None.
        """
        raise NotImplementedError

    @staticmethod
    async def set_event(event: Event):
        """
        :param event: The event to set
        """
        event.set()

    def deploy_together(self, deployment_dict: Dict[AgentReservation, Dict[str, Any]],
                        source: str,
                        max_lifetime_in_seconds: float):
        """
        :param deployment_dict: A dictionary whose keys are AgentReservations
                            and whose values are agent network spec dictionaries.
        :param source: A string describing where the deployment was coming from
        :param max_lifetime_in_seconds: Longest a temp network can live
        """
        # Get a timestamp as a basis of coordinated expiration.
        use_now: float = time()

        reservations_dict: Dict[Reservation, Dict[str, Any]] = {}
        for reservation, agent_spec in deployment_dict.items():

            # For each reservation in the block, set a consistent expiration time
            reservation.set_expiration_from(use_now, max_lifetime_in_seconds)

            # Put together a dictionary of reservations -> agent specs
            reservations_dict[reservation] = agent_spec

        # Deploy locally.
        for storage in self.reservations_storage:
            # This can synchronously lock
            storage.add_reservations(reservations_dict, source)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator


class CompositeDictionaryValidator(DictionaryValidator):
    """
    Implementation of the DictionaryValidator interface that uses multiple validators
    """

    def __init__(self, validators: List[DictionaryValidator]):
        """
        Constructor

        :param validators: A list of validators to use
        """
        self.validators: List[DictionaryValidator] = validators

    def validate(self, candidate: Dict[str, Any]) -> List[str]:
        """
        Validate the agent network.

        :param candidate: The dictionary to validate
        :return: A list of error messages
        """
        errors: List[str] = []

        if not candidate:
            errors.append("Nothing to validate.")
            return errors

        if self.validators is None or len(self.validators) == 0:
            errors.append("No validation policy.")
            return errors

        for validator in self.validators:
            errors.extend(validator.validate(candidate))

        return errors




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Tuple

from logging import getLogger
from logging import Logger

import json
import os
import aiofiles

from openfga_sdk import ClientConfiguration
from openfga_sdk.credentials import CredentialConfiguration
from openfga_sdk.credentials import Credentials
from openfga_sdk.exceptions import FgaValidationException
from openfga_sdk.models.create_store_request import CreateStoreRequest
from openfga_sdk.models.create_store_response import CreateStoreResponse
from openfga_sdk.models.list_stores_response import ListStoresResponse
from openfga_sdk.models.read_authorization_models_response import ReadAuthorizationModelsResponse
from openfga_sdk.models.write_authorization_model_request import WriteAuthorizationModelRequest
from openfga_sdk.models.write_authorization_model_response import WriteAuthorizationModelResponse
from openfga_sdk.client.client import OpenFgaClient


class OpenFgaInit:
    """
    Class to initialize the OpenFGA client where:
        * an OpenFGA store will be created if one does not exist
        * an authorization policy will be uploaded to that store if one does not exist.

    Per the Open FGA docs here:
        https://github.com/openfga/python-sdk?tab=readme-ov-file#initializing-the-api-client

        "We strongly recommend you initialize the OpenFgaClient only once
         and then re-use it throughout your app..."

    ... so this is not to be used by workaday client code.
    Use OpenFgaClientCache.get() instead.
    """

    def __init__(self):
        """
        Constructor.

        This is not to be used by workaday client code.
        Use OpenFgaClientCache.get() instead.
        """
        self.logger: Logger = getLogger(self.__class__.__name__)

    async def initialize_store(self, store_name: str) -> str:
        """
        Initialize a store for general usage.
        This might involve initializing a whole bunch of other stuff if the OpenFGA
        infrastructure has not yet been set up and so is expected to take awhile
        the first time around for any given store_name.
        """
        store_id: str = None
        open_fga_client: OpenFgaClient = self.initialize_one_client(store_id=store_id)
        async with open_fga_client as client:
            open_fga_client, store_id = await self.maybe_initialize_store(client, store_name)

        async with open_fga_client as client:
            await self.prepare_policy(client)
            await self.sync(client)

        return store_id

    @staticmethod
    def initialize_one_client(store_id: str = None, model_id: str = None) -> OpenFgaClient:
        """
        Initializes an OpenFgaClient.  Logic from:
            https://openfga.dev/docs/getting-started/setup-sdk-client

        :param store_id: A specific fact store_id for the client to connect to.
                         Default is None and certain client operations like CreateStore and ListStores
                         do not need a store_id.
        :param authorization_model_id: A default authorization model for the client to use within the store.
                        Can be None. This is OK in certain initialization situations.
                        Many client API calls can have the model_id as part of its request,
                        but to workaday calling code it's much more conventient to have this settled.
        """

        # Get OpenFGA env vars that point to its cloud service.
        # This might also want to come from setup config eventually.
        api_url: str = os.environ.get("FGA_API_URL")
        api_token: str = os.environ.get("FGA_API_TOKEN")

        credentials = None
        if api_token is not None:
            # Checkmarx flags this as a source for Privacy Violation path 2, 3, 4
            # This is a False Positive. Credentials themselves are not leaked anywhere.
            credentials = Credentials(method="api_token",
                                      configuration=CredentialConfiguration(api_token=api_token))
        configuration = ClientConfiguration(api_url=api_url,
                                            store_id=store_id,
                                            authorization_model_id=model_id,
                                            credentials=credentials)

        # Initialize our OpenFGA connection
        fga_client = OpenFgaClient(configuration)

        return fga_client

    async def maybe_initialize_store(self, open_fga_client: OpenFgaClient, store_name: str) \
            -> Tuple[OpenFgaClient, str]:
        """
        Determines whether or not the fact store needs one-time initialization
        from ground-zero.  Logic from:
            https://openfga.dev/docs/getting-started/create-store

        :param open_fga_client: An OpenFgaClient to make the determination about the store
        :param store_name: The name of the store
        :return: Same or different OpenFgaClient that is initialized to talk to the store.
        """

        # Look for our store
        use_store_name: str = os.environ.get("TEST_FGA_STORE_NAME", store_name)
        self.logger.info("Using store name %s", use_store_name)
        store_id: str = None

        # This is the first place that we attempt to connect to the OpenFGA server.
        try:
            response: ListStoresResponse = await open_fga_client.list_stores()

        except FgaValidationException as exception:
            # Make sure the error reflects what is going on better
            raise ValueError("FGA_API_URL env var not set correctly") from exception

        for store in response.stores:
            # store is a openfga_sdk.models.store.Store
            if store.name == use_store_name:
                store_id = store.id
                # Checkmarx flags this as a destination for Privacy Violation path 4
                # This is a False Positive. store ids and names themselves are not actually sensitive information.
                self.logger.info("Found %s for name %s", store_id, use_store_name)
                break

        # Create the store and a new client if not found
        if not store_id:
            body = CreateStoreRequest(name=use_store_name)
            response: CreateStoreResponse = await open_fga_client.create_store(body)
            store_id: str = response.id
            # Checkmarx flags this as a destination for Privacy Violation path 3
            # This is a False Positive. "store_id" itself is not actually sensitive information.
            self.logger.info("Created FGA store id %s", store_id)

        # From here on, always create a new client with the store_id configured
        new_client: OpenFgaClient = self.initialize_one_client(store_id)

        return new_client, store_id

    async def prepare_policy(self, open_fga_client: OpenFgaClient):
        """
        Prepare authorization policy for upload
        By the end of this method the existing open_fga_client member
        will be modified to use the current policy.
        """

        open_fga_policy_file: str = os.environ.get("FGA_POLICY_FILE", "")
        if not open_fga_policy_file:
            # We expect a json containing the auth model policy from the following command:
            # fga model transform --file ${POLICY_FILE_ROOT}.fga | python -m json.tool > ${POLICY_FILE_ROOT}.json
            raise ValueError("FGA_POLICY_FILE env var not set")

        # Read the OpenFGA policy from configuration
        policy: Dict[str, Any] = {}
        async with aiofiles.open(open_fga_policy_file, "r", encoding="utf-8") as policy_file:
            content: str = await policy_file.read()
            policy = json.loads(content)

        found_auth_model: str = await self.find_auth_model(open_fga_client, policy)

        # Do we need to store a new version of the auth model?
        if found_auth_model is None:
            found_auth_model = await self.update_auth_model(open_fga_client, policy)

        # For now. Needs more to set the model up
        # Checkmarx flags this as a destination for Privacy Violation path 1 and 2
        # This is a False Positive. Any auth model id itself is not actually sensitive information.
        self.logger.info("FGA auth model id %s", found_auth_model)

    async def find_auth_model(self, open_fga_client: OpenFgaClient, policy: Dict[str, Any]) -> str:
        """
        Looks for an existing authorization model that matches the
        model_info dictionary in the policy file.
        :param policy: The dictionary from the auth model JSON file
        :return: The AuthorizationModel id for the store that matches the version and
                checksum in the model_info.  If no such model exists, then None
                is returned.
        """
        _ = policy

        # Ask about existing auth models associated with the store
        response: ReadAuthorizationModelsResponse = await open_fga_client.read_authorization_models()

        # Look for the authorization model we specified.
        found_auth_model: str = None

        # Checkmarx flags this as a source for Privacy Violation path 1
        # This is a False Positive. Any auth model id itself is not actually sensitive information.
        auth_model_id: str = os.environ.get("FGA_MODEL_ID")

        # Find the model_info in the json file.
        for auth_model in response.authorization_models:

            # Look for what was specified in the env var (if anything was specified there)
            if auth_model.id == auth_model_id:
                found_auth_model = auth_model.id
                break

            # Callout to allow for self-updating authorization servers
            if await self.is_seen_auth_model(open_fga_client, auth_model.id):
                found_auth_model = auth_model.id

        return found_auth_model

    async def update_auth_model(self, open_fga_client: OpenFgaClient, policy: Dict[str, Any]) -> str:
        """
        Updates the auth model and associates it with the store.
        :param policy: The dictionary from the auth model JSON file
        :return: The AuthorizationModel id for the new model created for the store
        """
        auth_model_id: str = None

        # Find the model_info in the json file.
        model_info: Dict[str, Any] = {}
        model_info = policy.get("model_info", model_info)

        request = WriteAuthorizationModelRequest(
            type_definitions=policy.get("type_definitions"),
            schema_version=policy.get("schema_version"),
            conditions=policy.get("conditions"),
            local_vars_configuration=policy.get("local_vars_configuration"))
        response: WriteAuthorizationModelResponse = \
            await open_fga_client.write_authorization_model(request)
        auth_model_id = response.authorization_model_id

        # Update the model version and checksum now that we've written the model itself.
        # One of the few methods that is not async.
        open_fga_client.set_authorization_model_id(auth_model_id)

        return auth_model_id

    @staticmethod
    async def remove_store_for_testing(open_fga_client: OpenFgaClient):
        """
        Removes the store associated with the given client.
        :param open_fga_client: The client whose store is to be deleted
        """
        if open_fga_client is not None:
            await open_fga_client.delete_store()

    async def sync(self, open_fga_client: OpenFgaClient) -> None:
        """
        Optional method to synchronize the OpenFGA server with any new model changes.
        """
        # Do nothing, but allow overrides for specific circumstances.
        _ = open_fga_client

    async def is_seen_auth_model(self, open_fga_client: OpenFgaClient, auth_model_id: str) -> bool:
        """
        Optional method to allow for self-updating authorization servers
        :param auth_model_id: The auth model id to look for
        :return: True if the auth model was found
        """
        _ = open_fga_client, auth_model_id
        return False




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

import os

from pathlib import Path

from leaf_common.config.dictionary_overlay import DictionaryOverlay
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san import TOP_LEVEL_DIR
from neuro_san.internals.graph.activations.branch_activation import BranchActivation
from neuro_san.internals.graph.activations.class_activation import ClassActivation
from neuro_san.internals.graph.activations.external_activation import ExternalActivation
from neuro_san.internals.graph.activations.front_man_activation import FrontManActivation
from neuro_san.internals.graph.activations.sly_data_redactor import SlyDataRedactor
from neuro_san.internals.graph.activations.toolbox_activation import ToolboxActivation
from neuro_san.internals.graph.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.graph.interfaces.callable_activation import CallableActivation
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.front_man import FrontMan
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing


class ActivationFactory(AgentToolFactory):
    """
    A factory class for creating Activations of tools within the agent network graph.
    That is, this is where neuro-san tools are made real.
    """

    def __init__(self, agent_network: AgentNetwork):
        """
        Constructor

        :param agent_network: The AgentNetwork this factory will be basing its information on
        """
        self.agent_network: AgentNetwork = agent_network
        self.agent_tool_path: str = self._determine_agent_tool_path()

    def _determine_agent_tool_path(self) -> str:
        """
        Policy for determining where tool source should be looked for
        when resolving references to coded tools.

        :return: the agent tool path to use for source resolution.
        """
        # Try the env var first if nothing to start with
        agent_tool_path: str = os.environ.get("AGENT_TOOL_PATH")

        # Try reach-around directory if still nothing to start with
        if agent_tool_path is None:
            agent_tool_path = TOP_LEVEL_DIR.get_file_in_basis("coded_tools")

        # If we are dealing with file paths, convert that to something resolvable
        if agent_tool_path.find(os.sep) >= 0:

            # Find the best of many resolution paths in the PYTHONPATH
            resolved_tool_path: str = str(Path(agent_tool_path).resolve())
            best_path = ""
            pythonpath: str = os.environ.get("PYTHONPATH")
            if pythonpath is None:
                # Trust what we have already
                best_path = agent_tool_path
            else:
                pythonpath_split = pythonpath.split(os.pathsep)
                for one_path in pythonpath_split:
                    resolved_path: str = str(Path(one_path).resolve())
                    if resolved_tool_path.startswith(resolved_path) and \
                            len(resolved_path) > len(best_path):
                        best_path = resolved_path

            if len(best_path) == 0:
                raise ValueError(f"No reasonable agent tool path found in PYTHONPATH for {agent_tool_path}")

            # Find the path beneath the python path
            path_split = resolved_tool_path.split(best_path)
            if len(path_split) < 2:
                raise ValueError("""
Cannot find tool path for {agent_tool_path} in PYTHONPATH.
Check to be sure your value for PYTHONPATH includes where you expect where your coded tools live.
""")
            resolve_path = path_split[1]

            # Replace separators with python delimiters for later resolution
            agent_tool_path = resolve_path.replace(os.sep, ".")

            # Remove any leading .s
            while agent_tool_path.startswith("."):
                agent_tool_path = agent_tool_path[1:]

        # Now, agent network name itself can contain "/" symbols (regardless of underlying OS)
        # in case of hierarchical agents structure. Replace those with "." as well.
        agent_network_path = self.agent_network.get_network_name().replace("/", ".")

        # Be sure the name of the agent (stem of the hocon file) is the
        # last piece to narrow down the path resolution further.
        if not agent_tool_path.endswith(agent_network_path):
            agent_tool_path = f"{agent_tool_path}.{agent_network_path}"

        return agent_tool_path

    def get_agent_tool_path(self) -> str:
        """
        :return: The path under which tools for this registry should be looked for.
        """
        return self.agent_tool_path

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def create_agent_activation(self, parent_run_context: RunContext,
                                parent_agent_spec: Dict[str, Any],
                                name: str,
                                sly_data: Dict[str, Any],
                                arguments: Dict[str, Any] = None,
                                factory: AgentToolFactory = None,
                                invocation: str = None) -> CallableActivation:
        """
        Create an active node for an agent from its spec.
        This is how CallableActivations create other CallableActivations.

        :param parent_run_context: The RunContext of the agent calling this method
        :param parent_agent_spec: The spec of the agent calling this method.
        :param name: The name of the agent to get out of the registry
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        :param arguments: A dictionary of arguments for the newly constructed agent
        :param factory: A factory that will be used to create the agent tool
        :param invocation: The invocation style of the activation.
        :return: The CallableActivation agent referred to by the name.
        """
        if factory is None:
            factory = self

        agent_activation: CallableActivation = None

        agent_tool_spec: Dict[str, Any] = self.agent_network.get_agent_tool_spec(name)
        if agent_tool_spec is None:

            if not ExternalAgentParsing.is_external_agent(name):
                raise ValueError(f"No agent_tool_spec for {name}")

            # For external tools, we want to redact the sly data based on
            # the calling/parent's agent specs.
            redacted_sly_data: Dict[str, Any] = self._redact_sly_data(parent_run_context, sly_data)

            # Get the spec for allowing upstream data
            extractor = DictionaryExtractor(parent_agent_spec)
            empty = {}
            allow_from_downstream: Dict[str, Any] = extractor.get("allow.from_downstream", empty)

            agent_activation = ExternalActivation(parent_run_context, factory, name, arguments, redacted_sly_data,
                                                  allow_from_downstream, invocation)
            return agent_activation

        # Merge the arguments coming in from the LLM with those that were specified
        # in the hocon file for the agent.
        use_args: Dict[str, Any] = self._merge_args(arguments, agent_tool_spec)

        if agent_tool_spec.get("toolbox") is not None:
            # If a toolbox is in the spec, this is a shared coded tool where tool's description and
            # args schema are defined in either AGENT_TOOLBOX_INFO_FILE or toolbox_info.hocon.
            agent_activation = ToolboxActivation(parent_run_context, factory, use_args, agent_tool_spec, sly_data)
            return agent_activation

        if agent_tool_spec.get("function") is not None:
            # If we have a function in the spec, the agent has arguments
            # it wants to be called with.
            if agent_tool_spec.get("class") is not None:
                # Agent specifically requested a python class to be run,
                # and tool's description and args schema are defined in agent network hocon.
                agent_activation = ClassActivation(parent_run_context, factory, use_args, agent_tool_spec, sly_data)
            else:
                agent_activation = BranchActivation(parent_run_context, factory, use_args, agent_tool_spec, sly_data)
        else:
            # Get the tool to call from the spec.
            agent_activation = FrontManActivation(parent_run_context, factory, agent_tool_spec, sly_data)

        return agent_activation

    def create_front_man(self,
                         sly_data: Dict[str, Any] = None,
                         parent_run_context: RunContext = None,
                         factory: AgentToolFactory = None) -> FrontMan:
        """
        Find and create the FrontMan for DataDrivenChat

        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        :param parent_run_context: A RunContext instance
        :param factory: An optional extra parameter at this ActivationFactory level to provide
                    the correct object reference for factory scope/lifetime issues.
        """
        if factory is None:
            factory = self

        front_man_name: str = self.agent_network.find_front_man()

        agent_tool_spec: Dict[str, Any] = self.agent_network.get_agent_tool_spec(front_man_name)
        front_man = FrontManActivation(parent_run_context, factory, agent_tool_spec, sly_data)
        return front_man

    def get_config(self) -> Dict[str, Any]:
        """
        :return: The entire config dictionary given to the instance.
        """
        return self.agent_network.get_config()

    def get_name_from_spec(self, agent_spec: Dict[str, Any]) -> str:
        """
        :param agent_spec: A single agent to register
        :return: The agent name as per the spec
        """
        return self.agent_network.get_name_from_spec(agent_spec)

    def _merge_args(self, llm_args: Dict[str, Any], agent_tool_spec: Dict[str, Any]) -> Dict[str, Any]:
        """
        Merges the args specified by the llm with "hard-coded" args specified in the agent spec.
        Hard-coded args win over llm-specified args if both are defined.
        If you want the llm args to win out over the hard-coded args, use a default for
        the function spec instead of the hard-coded args.

        :param llm_args: argument dictionary that the LLM wants
        :param agent_tool_spec: The dictionary representing the spec registered agent
        """
        config_args: Dict[str, Any] = agent_tool_spec.get("args")
        if config_args is None:
            # Nothing to override
            return llm_args

        overlay = DictionaryOverlay()
        merged_args: Dict[str, Any] = overlay.overlay(llm_args, config_args)
        return merged_args

    def _redact_sly_data(self, parent_run_context: RunContext, sly_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Redact the sly_data based on the agent spec associated with the parent run context

        :param parent_run_context: The parent run context of the tool to be created.
        :param sly_data: The internal representation of the sly_data to be redacted
        :return: A new sly_data dictionary, redacted as per the parent spec
        """
        parent_spec: Dict[str, Any] = None
        if parent_run_context is not None:
            parent_spec = parent_run_context.get_agent_tool_spec()

        redactor = SlyDataRedactor(parent_spec, config_keys=["allow.sly_data", "allow.to_downstream.sly_data"])
        redacted: Dict[str, Any] = redactor.filter_config(sly_data)
        return redacted




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.messages.chat_message_type import ChatMessageType


class MessageFilter:
    """
    An interface for filtering a single message.
    """

    def allow(self, chat_message_dict: Dict[str, Any]) -> bool:
        """
        Determine whether to allow the message through.
        This is the main entry point for most clients.
        """
        message_type: ChatMessageType = self.get_message_type(chat_message_dict)
        return self.allow_message(chat_message_dict, message_type)

    @staticmethod
    def get_message_type(chat_message_dict: Dict[str, Any]) -> ChatMessageType:
        """
        Convert the message type in the ChatMessage dictionary to the enum we want to work with
        :param chat_message_dict: The ChatMessage dictionary to consider.
        :return: The ChatMessageType of the chat_message_dict
        """
        response_type: str = chat_message_dict.get("type")
        message_type: ChatMessageType = ChatMessageType.from_response_type(response_type)
        return message_type

    def allow_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType) -> bool:
        """
        Determine whether to allow the message through.
        This is what subclasses should implement.
        MessageProcessors also call this interface directly.

        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        :return: True if the message should be allowed through to the client. False otherwise.
        """
        raise NotImplementedError
